﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HeuristicLab.Algorithms.Bandits {
  public class TruncatedNormalBandit {
    public int NumArms { get; private set; }
    public double OptimalExpectedReward { get; private set; } // reward of the best arm, for calculating regret
    private readonly Random random;
    private readonly double[] expReward;
    public TruncatedNormalBandit(Random random, int nArms) {
      this.random = random;
      this.NumArms = nArms;
      // expected reward of arms is iid and uniformly distributed 
      expReward = new double[nArms];
      OptimalExpectedReward = double.NegativeInfinity;
      for (int i = 0; i < nArms; i++) {
        expReward[i] = random.NextDouble();
        if (expReward[i] > OptimalExpectedReward) OptimalExpectedReward = expReward[i];
      }
    }

    // pulling an arm results in a truncated normally distributed reward 
    // with mean expReward[i] and std.dev 0.1
    public double Pull(int arm) {
      double x = 0;
      do {
        var z = Transform(random.NextDouble(), random.NextDouble());
        x = z * 0.1 + expReward[arm];
      }
      while (x < 0 || x > 1);
      return x;
    }

    // box muller transform
    private double Transform(double u1, double u2) {
      return Math.Sqrt(-2 * Math.Log(u1)) * Math.Cos(2 * Math.PI * u2);
    }
  }
}
