﻿using System;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.GrammaticalOptimization {
  public class RandomSearch {
    public event Action<string, double> FoundNewBestSolution;
    public event Action<string, double> SolutionEvaluated;

    private readonly int maxLen;
    private readonly Random random;

    public RandomSearch(int maxLen) {
      this.maxLen = maxLen;
      this.random = new Random(31415);
    }

    public void Run(IProblem problem, int maxIterations) {
      double bestQuality = double.MinValue;
      for (int i = 0; i < maxIterations; i++) {
        var sentence = CreateSentence(problem.Grammar);
        var quality = problem.Evaluate(sentence);
        RaiseSolutionEvaluated(sentence, quality);

        if (quality > bestQuality) {
          bestQuality = quality;
          RaiseFoundNewBestSolution(sentence, quality);
        }
      }
    }

    private string CreateSentence(IGrammar grammar) {
      var sentence = grammar.SentenceSymbol.ToString();
      return grammar.CompleteSentenceRandomly(random, sentence, maxLen);
    }

    private void RaiseSolutionEvaluated(string sentence, double quality) {
      var handler = SolutionEvaluated;
      if (handler != null) handler(sentence, quality);
    }
    private void RaiseFoundNewBestSolution(string sentence, double quality) {
      var handler = FoundNewBestSolution;
      if (handler != null) handler(sentence, quality);
    }
  }
}
